{% load humanize %}
{% load evelinks %}

<li class="list-group-item">
    <img src="{{ asset.eve_type_id|type_icon_url:128 }}" style="height: 32px; width: 32px;" title="{{ asset.eve_type.name }}">&nbsp;
    {% if asset.eve_type.profile_url %}
        <a href="{{ asset.eve_type.profile_url }}" target="_blank">{{ asset.eve_type.name }}</a>
    {% else %}
        {{ asset.eve_type.name }}
    {% endif %}
    {% if asset.is_singleton is False %}
        <span class="pull-right">{{ asset.quantity|default_if_none:"?"|intcomma }}</span>
    {% endif %}
</li>
