{% load i18n %}
{% load bootstrap %}

<div role="tabpanel" class="tab-pane" id="summary">
    <div class="table-responsive">
        <table id="tab_summary" style="width: 100%" class="table table-striped table-compact">
            <thead>
                <tr>
                    <th></th>
                    <th>{% translate "Corporation" %}</th>
                    <th>{% translate "Alliance" %}</th>
                    <th class="text-right">{% translate "Citadels" %}</th>
                    <th class="text-right">{% translate "Engineering Complexes" %}</th>
                    <th class="text-right">{% translate "Refineries" %}</th>
                    <th class="text-right">{% translate "Other Upwell" %}</th>
                    <th class="text-right">{% translate "Customs Offices" %}</th>
                    <th class="text-right">{% translate "Starbases" %}</th>
                    <th class="text-right">{% translate "Total" %}</th>
                </tr>
            </thead>
            <tfoot>
                <tr class="info">
                    <th></th>
                    <th>Total</th>
                    <th></th>
                    <th class="text-right"></th>
                    <th class="text-right"></th>
                    <th class="text-right"></th>
                    <th class="text-right"></th>
                    <th class="text-right"></th>
                    <th class="text-right"></th>
                    <th class="text-right"></th>
                </tr>
            </tfoot>
        </table>
    </div>
</div>
