# Generated by Django 3.1.13 on 2021-10-05 17:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("structures", "0030_update_refueled_notifications"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="owner",
            name="assets_last_update_ok",
        ),
        migrations.RemoveField(
            model_name="owner",
            name="forwarding_last_update_ok",
        ),
        migrations.RemoveField(
            model_name="owner",
            name="notifications_last_update_ok",
        ),
        migrations.RemoveField(
            model_name="owner",
            name="structures_last_update_ok",
        ),
        migrations.AddField(
            model_name="owner",
            name="is_up",
            field=models.BooleanField(
                default=None,
                editable=False,
                help_text="whether all services for this owner are currently up",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="assets_last_update_at",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="when the last successful update happened",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="forwarding_last_update_at",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="when the last successful update happened",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="notifications_last_update_at",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="when the last successful update happened",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="owner",
            name="structures_last_update_at",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="when the last successful update happened",
                null=True,
            ),
        ),
    ]
